##############################################################################
#
# Copyright (c) 2002 Zope Corporation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.0 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE
#
##############################################################################
"""ZCatalog text index for indexing multiple fields of objects

$Id: index.py,v 1.11 2004/01/20 04:13:36 caseman Exp $"""

from Globals import DTMLFile, InitializeClass
import Acquisition
from Persistence import Persistent
from OFS.SimpleItem import SimpleItem
from Products.PluginIndexes.common.util import parseIndexRequest
from AccessControl import Permissions
from AccessControl.SecurityInfo import ClassSecurityInfo
from Products.ZCTextIndex.QueryParser import QueryParser
from Products.ZCTextIndex import WidCode
from Products.ZCTextIndex.OkapiIndex import OkapiIndex
from Products.ZCTextIndex.BaseIndex \
    import BaseIndex, inverse_doc_frequency, scaled_int, unique
from Products.ZCTextIndex.SetOps import mass_weightedIntersection
from Products.ZCTextIndex.okascore import score
from BTrees.IOBTree import IOBTree
from BTrees.IIBTree import IIBTree, IITreeSet, IIBucket, IISet
from BTrees.IIBTree \
    import weightedIntersection, multiunion, difference, intersection

try:
    from Products.PluginIndexes.common.PluggableIndex \
        import UniqueValueIndex
except ImportError:
    # Older Zope (pre 2.7) w/o UniqueValueIndex
    from Products.PluginIndexes.common.PluggableIndex \
        import PluggableIndexInterface as UniqueValueIndex

class FieldedTextIndex(OkapiIndex, SimpleItem):
    """Multiple field text index"""

    __implements__ = UniqueValueIndex
    
    meta_type = 'Fielded Text Index'

    manage_options = (
        {'label': 'Overview', 'action': 'manage_main'},
        {'label': 'Default Field Weights', 'action':'manage_weights'},
    )
    
    manage_main = DTMLFile('www/manageIndex', globals())
    manage_weights = DTMLFile('www/manageWeights', globals())
    
    query_options = ['query', 'fields', 'field_weights']
    
    security = ClassSecurityInfo()
    security.declareObjectProtected(Permissions.manage_zcatalog_indexes)
    
    def __init__(
        self, id, source_name=None, lexicon=None, extra=None, caller=None, ):
        """Fielded text index constructor
        
        id -- Zope id for index in
        
        source_name -- Name of attribute on indexed objects that contains
        or returns the index source data
        
        lexicon -- ZCTextIndex Lexicon object used by the index
        
        extra and caller are used by the wonderous ZCatalog addIndex 
        machinery. You can ignore them, unfortunately I can't 8^/
        """
        self.id = id
        self.source_name = source_name or extra.source_name
        OkapiIndex.__init__(self, None)
        
        if lexicon is not None:
            # Yes, we are storing a hard reference to the Lexicon
            # and so does ZCTextIndex internally. Acquisition-be-damned
            self._lexicon = lexicon
        else:
            # acquire the lexicon from the caller (the ZCatalog)
            self._lexicon = getattr(caller, extra.lexicon_id)
        
        # Mapping wid -> fields -> docids
        self._wordfields = IOBTree()
        
        # List of field names for this index. The index of the name in
        # this list is the fieldid used in the index data structures
        # This can be a simple list since I don't anticipate a large 
        # number of fields (probably several hundred max) and it will be
        # changed infrequently
        self._fields = []
        self._default_weights = {} # map weight => fieldids
    
    ## Index specific methods ##
    
    security.declareProtected(Permissions.search_zcatalog, 'fieldNames')
    def fieldNames(self):
        """Return a list of the field names in the index"""
        return self._fields[:]
    
    security.declareProtected(Permissions.search_zcatalog, 'lexiconId')
    def lexiconId(self):
        """Return the id of the lexicon for this index"""
        return self._lexicon.getId()
        
    security.declareProtected(
        Permissions.manage_zcatalog_indexes, 'setDefaultWeights')
    def setDefaultWeights(self, weights):
        """Set the default weights where weights is a mapping of
        field name => integer weight. weights may also be a record object
        for convenience when setting from a form. Note that fields can be
        assigned weights before objects with those fields have been indexed.
        """
        self._default_weights = self._mapWeights2Fields(weights)
    
    def _mapWeights2Fields(self, weights):
        weights2fields = {}
        for fieldid, fname in enumerate(self._fields):                
            weight = int(weights.get(fname, 1))
            if weight in weights2fields:
                weights2fields[weight].insert(fieldid)
            else:
                weights2fields[weight] = IISet((fieldid,))
        return weights2fields
    
    security.declareProtected(Permissions.search_zcatalog, 'getDefaultWeights')
    def getDefaultWeights(self):
        """Return a dictionary of the default weights for every field"""
        defaults = {}
        for weight, fieldids in self._default_weights.items():
            for fid in fieldids:
                defaults[self._fields[fid]] = weight
        return defaults
    
    ## Pluggable Index API ##
    
    def index_object(self, docid, obj, theshold=None):
        """Add document to index"""
        source = getattr(obj, self.source_name, None)
        if source is not None:
            # Try calling source
            try:
                source = source()
            except TypeError:
                pass
            # Try using the mapping protocol to get pairs
            try:
                source = source.items()
            except (TypeError, AttributeError):
                pass
        self.unindex_object(docid)
        if not source:
            return 0
        frequencies = {}
        freq_get = frequencies.get
        lexicon = self._lexicon
        field_wids = {}
        # Get wids for each field and total word frequencies for the whole doc
        for fname, text in source:
            fieldid = self._field_id(fname)
            field_wids[fieldid] = wids = lexicon.sourceToWordIds(text)
            for wid in wids:
                frequencies[wid] = freq_get(wid, 0) + 1
        self._mass_add_wordinfo(frequencies, docid)
        
        field2words = {}
        # Add wordinfo for each field
        for fieldid, wids in field_wids.items():
            for wid in wids:
                try:
                    fielddocs = self._wordfields[wid]
                except KeyError:
                    self._wordfields[wid] = fielddocs = IOBTree()
                    fielddocs[fieldid] = wordids = IITreeSet()
                else:
                    try:
                        wordids = fielddocs[fieldid]
                    except KeyError:
                        fielddocs[fieldid] = wordids = IITreeSet()
                wordids.insert(docid)
            field2words[fieldid] = WidCode.encode(wids)
        
        # Add doc info
        self._docweight[docid] = len(frequencies)
        self._docwords[docid] = field2words # map: fieldid -> compressed wids
        self.document_count.change(1)
        count = len(frequencies)
        self._change_doc_len(count)
        return count
        
    '''
    def _reindex_object(self, docid, source):
        """Reindex an existing document taking care to actual update only
        the fields and words that have changed"""
        field2words = self._docwords[docid]
        for fname, text in source:
            fieldid = self._field_id(fname)
            field_wids[fieldid] = wids = lexicon.sourceToWordIds(text)
            encoded_wids = WidCode.encode(wids)
            if encoded_wids != field2words.get(fieldid):
                # This field has changed
                new_wids = IITreeSet(wids)
                try:
                    old_widcodes = field2words[fieldid]
                except KeyError:
                    # New field for this doc
                    self._mass_add_
                else:
                    old_wids = IITreeSet(WidCode.decode(old_widcodes))
                    wids_added = difference(new_wids, old_wids)
                    wids_removed = difference(old_wids, new_wids)
    '''
    
    def unindex_object(self, docid):
        """Remove docid from the index"""
        try:
            field2words = self._docwords[docid]
        except KeyError:
            return 0 # docid not in index
        for fieldid, words in field2words.items():
            for wid in WidCode.decode(words):
                self._del_wordinfo(wid, fieldid, docid)
        del self._docwords[docid]
        del self._docweight[docid]
        self.document_count.change(-1)
        return 1      
    
    def _del_wordinfo(self, wid, fieldid, docid):
        # Override from BaseIndex
        try:
            doc2score = self._wordinfo[wid]
            del doc2score[docid]
        except KeyError:
            pass
        else:
            if not doc2score:
                del self._wordinfo[wid]
                self.length.change(-1)
        try:
            fielddocs = self._wordfields[wid][fieldid]
            fielddocs.remove(docid)
        except KeyError:
            pass
        else:
            if not fielddocs:
                del self._wordfields[wid]
    
    def _apply_index(self, request, cid=''):
        """Apply query specified by request, a mapping containing the query.
        
        Queries are in the form of::
            
          {'query':'words to search', 'fields':['foobar', 'splat', 'bizbaz']}
        
        fields limits the query to only the field names specified. Unknown 
        fields are ignored. If fields is omitted, then all fields are searched.

        Returns two object on success, the resultSet containing the
        matching record numbers and a tuple containing the names of
        the fields used

        Returns None if request is not valid for this index.
        """
        record = parseIndexRequest(
            request, self.source_name, self.query_options)
        if record.keys is None:
            return None
        query_str = ' '.join(record.keys)
        if not query_str:
            return None
        fields = record.get('fields')
        field_weights = record.get('field_weights')
        
        if fields:                
            # Get the search field, omitting ones we don't know about
            # This avoids errors in legitimate app code searching a catalog
            # not yet populated with objects containing all expected fields
            fieldids = IISet()
            for fname in fields:
                try:
                    fid = self._fields.index(fname)
                except ValueError:
                    pass
                else:
                    fieldids.insert(fid)
            
            if not fieldids:
                # There were no fields we have indexed
                return IIBucket(), (self.source_name,)
        else:
            fieldids = None
            
        if field_weights is not None:
            weights2fields = self._mapWeights2Fields(field_weights)
        else:
            weights2fields = self._default_weights
                                
        # Because the parse tree insulates us from the actual search code
        # pretty completely, we stuff fieldids in the REQUEST so that it
        # can be accessed by the _search_wids method down the line.
        # this is a hack, but it saves me from overriding the whole 
        # QueryParser and ParseTree (less code == good)
        self.REQUEST._fielded_text_index_options = (fieldids, weights2fields)
            
        tree = QueryParser(self._lexicon).parseQuery(query_str)
        results = tree.executeQuery(self)
        del self.REQUEST._fielded_text_index_options
        return  results, (self.source_name,)

    def getIndexSourceNames(self):
        return (self.source_name,)
        
    def getEntryForObject(self, docid, default=None):
        """Return a mapping of field -> words for docid"""
        try:
            field2words = self._docwords[docid]
        except KeyError:
            return default
        get_word = self._lexicon.get_word
        result = {}
        for fieldid, words in field2words.items():
            wids = WidCode.decode(words)
            result[self._fields[fieldid]] = [get_word(wid) for wid in wids]
        return result
        
    def hasUniqueValuesFor(self, name):
        return name == self.source_name
        
    def uniqueValues(self, name=None, withLengths=0):
        # Return the field names of everything we have indexed
        # withLengths is ignored
        if name is None or name == self.source_name:
            names = self.fieldNames()
            names.sort()
            return names
        else:
            return []
        
    ## ZCatalog ZMI methods ##

    def numObjects(self):
        """Return number of unique words in the index"""
        return self.length()

    def clear(self):
        """reinitialize the index (but not the lexicon)"""
        self.__init__(self.id, self.source_name, self._lexicon)
        
    ## Helper methods ##
    
    def _field_id(self, fname):
        """Return the field id for fname, assigning a new id if the name has
        not been seen before"""
        try:
            return self._fields.index(fname)
        except ValueError:
            # Encountered new field name
            # I expect this to rarely happen, so I am not concerned
            # about possible conflicts arising here
            self._fields.append(fname)
            fieldid = self._fields.index(fname)
            if 1 in self._default_weights:
                self._default_weights[1].insert(fieldid)
            else:
                self._default_weights[1] = IISet((fieldid,))
            self._p_changed = 1
            return fieldid
    
    def _mass_add_wordinfo(self, wid2weight, docid):
        # Override from BaseIndex
        dicttype = type({})
        get_doc2score = self._wordinfo.get
        new_word_count = 0
        for wid, weight in wid2weight.items():
            doc2score = get_doc2score(wid)
            if doc2score is None:
                self._wordinfo[wid] = doc2score = IIBTree()
                new_word_count += 1
            doc2score[docid] = weight
        self.length.change(new_word_count)
    
    def _search_wids(self, wids):
        """Override from OkapiIndex to search for words. Looks in the
        REQUEST for the fieldids to search (see _apply_index above)"""
        if not wids:
            return []
        search_fields, weights2fields = getattr(
            self.REQUEST, '_fielded_text_index_options', None)
        N = float(self.document_count())  # total # of docs
        doclen = self._totaldoclen()
        meandoclen = doclen / N
        #K1 = self.K1
        #B = self.B
        #K1_plus1 = K1 + 1.0
        #B_from1 = 1.0 - B

        #                           f(D, t) * (k1 + 1)
        #   TF(D, t) =  -------------------------------------------
        #               f(D, t) + k1 * ((1-b) + b*len(D)/E(len(D)))

        L = []
        docid2len = self._docweight
        for t in wids:
            d2f = self._wordinfo[t] # map {docid -> f(docid, t)}
            fielddocs = None
            for weight, fields in weights2fields.items():
                if search_fields is not None:
                    fields4weight = intersection(fields, search_fields)
                    if not fields4weight:
                        continue # Not searching any fields with this weight
                else:
                    fields4weight = fields # searching all fields
                if len(fields4weight) == len(self._fields):
                    # We are searching all fields indexed and they all have the
                    # same weight, so we don't need to weight them individually
                    idf = inverse_doc_frequency(len(d2f), N)
                    result = IIBucket()
                    score(result, d2f.items(), docid2len, idf, meandoclen)
                    L.append((result, weight))
                else:
                    if fielddocs is None:
                        try:
                            fielddocs = self._wordfields[t]
                        except KeyError:
                            # XXX I'm not sure _wordinfo should ever have a
                            # XXX key that _wordfields doesn't. This fixes
                            # XXX it for now at least
                            continue
                    docsets = []
                    for fieldid in fields4weight:
                        try:
                            docsets.append(fielddocs[fieldid])
                        except KeyError:
                            pass # No docs for this field
                    w, docs4weight = weightedIntersection(
                        d2f, multiunion(docsets), 1, 0)
                    if not docs4weight:
                        continue
                    idf = inverse_doc_frequency(len(docs4weight), N)
                    result = IIBucket()
                    score(result, docs4weight.items(), docid2len, 
                          idf, meandoclen)
                    L.append((result, weight))
        return L

    def search_phrase(self, phrase):
        """Override from BaseIndex to search fielded docwords mapping"""
        wids = self._lexicon.termToWordIds(phrase)
        cleaned_wids = self._remove_oov_wids(wids)
        if len(wids) != len(cleaned_wids):
            # At least one wid was OOV:  can't possibly find it.
            return IIBTree()
        scores = self._search_wids(wids)
        hits = mass_weightedIntersection(scores)
        if not hits:
            return hits
        code = WidCode.encode(wids)
        search_fields = getattr(
            self.REQUEST, '_fielded_text_index_search_fields', None)
        if search_fields is None:
            search_fields = range(len(self._fields))
        result = IIBTree()
        for docid, weight in hits.items():
            docwords = self._docwords[docid]
            for field in search_fields:
                try:
                    words = docwords[field]
                except KeyError:
                    pass # Word not found in this field for any doc
                else:
                    if words.find(code) >= 0:
                        result[docid] = weight
        return result
InitializeClass(FieldedTextIndex)

addIndexForm = DTMLFile('www/addIndex', globals())
